from shard.exceptions import NotFoundPoolException
from shard.utils.consistent_hash.pool import ConHashPool
from shard.utils.database import get_shard_groups, get_master_databases_by_shard_group, get_tables
from shard.utils.replica import get_replica_count


class PoolProvider:
    def __init__(self):
        self._pools = {}

    def create(self):
        shard_groups = get_shard_groups()

        for shard_group in shard_groups:
            databases = get_master_databases_by_shard_group(shard_group=shard_group)
            tables = get_tables(databases)
            self._pools[shard_group] = ConHashPool(nodes=tables, replica=get_replica_count(shard_group=shard_group))

    def get_pools(self, shard_group: str) -> ConHashPool:
        pool = self._pools.get(shard_group, None)
        if pool is None:
            raise NotFoundPoolException

        return pool

    def get_shard_index(self, shard_group: str, shard_key: int) -> int:
        pool = self.get_pools(shard_group=shard_group)
        return pool.get_node(value=shard_key).index


pool_provider = PoolProvider()



from shard.exceptions import NotShardingModelException, RequireShardKeyException
from shard.mixins import ShardMixin

__all__ = ('get_shard_key_from_kwargs', 'get_shard_key_from_instance', )


def get_shard_key_from_kwargs(model, **kwargs) -> int:
    _validate_shard_model(model)

    if model.shard_key_name not in kwargs:
        raise RequireShardKeyException()

    return kwargs[model.shard_key_name]


def get_shard_key_from_instance(instance: ShardMixin) -> int:
    return instance.get_shard_key()


def _validate_shard_model(model):
    if not issubclass(model, ShardMixin):
        raise NotShardingModelException()

    if model.shard_group is None or model.shard_key_name is None:
        raise NotShardingModelException()



class QueryFileHandler:
    @staticmethod
    def load_queries(file_name: str):
        file = open(file_name, 'r')
        query = file.read()
        file.close()

        query_list = [s and s.strip() for s in query.split(';')]
        return list(filter(None, query_list))



from django.core.management.base import BaseCommand
from multiprocessing.pool import Pool
from pprint import pprint
from typing import Dict, List

from shard.services.execute_query_service import ExecuteQueryService
from shard.services.query_file_handler import QueryFileHandler
from shard.utils.database import get_master_databases_by_shard_group


def run_process(params) -> Dict:
    if params['with_transaction']:
        result = ExecuteQueryService.execute_queries_by_shard_with_transaction(params['shard'], params['queries'])
    else:
        result = ExecuteQueryService.execute_queries_by_shard(params['shard'], params['queries'])

    return {
        'shard': params['shard'],
        'result': result
    }


class Command(BaseCommand):
    title = 'Execute the query on all shard.'

    def add_arguments(self, parser):
        parser.add_argument(
            '--shard_group', '-s',
            type=str,
            dest='shard_group',
            help='Shard group to execute',
            required=True,
        )

        parser.add_argument(
            '--sql_file', '-f',
            type=str,
            dest='sql_file',
            help='Path of file containing the query',
            required=True,
        )

        parser.add_argument(
            '--transaction', '-t',
            action='store_true',
            dest='with_transaction',
            help='Execute with transaction',
        )

    @staticmethod
    def print_queries(queries: List[str]):
        for query in queries:
            print(query)

    def handle(self, *args, **options):
        shard_group = options['shard_group']
        databases = get_master_databases_by_shard_group(shard_group)
        queries = QueryFileHandler.load_queries(options['sql_file'])
        params = [
            {'shard': database, 'queries': queries, 'with_transaction': options['with_transaction']} for database in databases
        ]

        self.print_queries(queries)

        pool = Pool(processes=len(databases))
        result = {
            'shard_group': shard_group,
            'result': pool.map(run_process, params),
        }

        pprint(result)





class ConHashNode:
    def __init__(self, hash_value: int, index: int):
        self.hash_value = hash_value
        self.index = index

    def __str__(self) -> str:
        return f"{self.index}:{self.hash_value}"



from hashlib import blake2b
from typing import Any


def blake2b_hash(value: Any) -> int:
    value = str(value)
    hash_value = int(blake2b(value.encode('utf-8')).hexdigest(), 16) % (2 ** 64)
    return hash_value



from django.core.management.base import BaseCommand
from pprint import pprint
from typing import List

from shard.services.execute_query_service import ExecuteQueryService
from shard.services.query_file_handler import QueryFileHandler


class Command(BaseCommand):
    title = 'Execute the query on a shard.'

    def add_arguments(self, parser):
        parser.add_argument(
            '--shard', '-s',
            type=str,
            dest='shard',
            help='Shard to execute the query',
            required=True,
        )

        parser.add_argument(
            '--sql_file', '-f',
            type=str,
            dest='sql_file',
            help='Path of file containing the query',
            required=True,
        )

        parser.add_argument(
            '--transaction', '-t',
            action='store_true',
            dest='with_transaction',
            help='Execute with transaction',
        )

    @staticmethod
    def print_queries(queries: List[str]):
        for query in queries:
            print(query)

    def handle(self, *args, **options):
        shard = options['shard']
        queries = QueryFileHandler.load_queries(options['sql_file'])

        self.print_queries(queries)

        if options['with_transaction']:
            result = ExecuteQueryService.execute_queries_by_shard_with_transaction(shard, queries)
        else:
            result = ExecuteQueryService.execute_queries_by_shard(shard, queries)

        result = {
            'shard': shard,
            'result': result
        }

        pprint(result)



from django.db import models

from shard.mixins import IsolatedShardMixin
from shard_static.managers import ShardStaticManager, ShardStaticStatusManager
from shard_static.mixins import ShardStaticMixin


class BaseStaticTransmitStatus(IsolatedShardMixin, models.Model):
    key = models.CharField(max_length=64, unique=True, verbose_name='Transmit Status Key')

    created = models.DateTimeField(auto_now_add=True, verbose_name='Created Datetime', )
    last_modified = models.DateTimeField(auto_now=True, verbose_name='Last Modified', )

    objects = ShardStaticStatusManager()

    class Meta:
        abstract = True

    @property
    def criterion(self):
        raise NotImplementedError

    @criterion.setter
    def criterion(self, criterion):
        raise NotImplementedError


class BaseShardStaticModel(ShardStaticMixin, models.Model):
    objects = ShardStaticManager()

    class Meta:
        abstract = True



# Generated by Django 2.0.2 on 2018-06-07 12:22

from django.db import migrations, models
import shard.mixins


class Migration(migrations.Migration):

    dependencies = [
        ('tests', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='StaticTransmitStatus',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('key', models.CharField(max_length=64, unique=True, verbose_name='Transmit Status Key')),
                ('created', models.DateTimeField(auto_now_add=True, verbose_name='Created Datetime')),
                ('last_modified', models.DateTimeField(auto_now=True, verbose_name='Last Modified')),
                ('criterion_id', models.IntegerField(default=1, verbose_name='Criterion ID')),
            ],
            options={
                'abstract': False,
            },
            bases=(shard.mixins.IsolatedShardMixin, models.Model),
        ),
    ]



from shard.managers import BaseShardManager
from shard_static.exceptions import NotExistsOriginalDataException


def _wrap_for_static(func_name):
    def wrapped(self, *args, **kwargs):
        if not self.model.transmit:
            raise NotExistsOriginalDataException()

        return getattr(self.get_queryset(), func_name)(*args, **kwargs)

    wrapped.__name__ = func_name
    return wrapped


def _wrap_for_status(func_name):
    def wrapped(self, shard: str, *args, **kwargs):
        return getattr(self.shard(shard=shard), func_name)(*args, **kwargs)

    wrapped.__name__ = func_name
    return wrapped


class ShardStaticStatusManager(BaseShardManager):
    filter = _wrap_for_status('filter')
    get = _wrap_for_status('get')
    create = _wrap_for_status('create')
    get_or_create = _wrap_for_status('get_or_create')
    update_or_create = _wrap_for_status('update_or_create')


class ShardStaticManager(BaseShardManager):
    filter = _wrap_for_static('filter')
    get = _wrap_for_static('get')
    create = _wrap_for_static('create')
    get_or_create = _wrap_for_static('get_or_create')
    update_or_create = _wrap_for_static('update_or_create')



from typing import Any, List

from shard.utils.consistent_hash import con_hash
from shard.utils.consistent_hash.node import ConHashNode

_MAX_32BIT_INTEGER = 2147483647


class ConHashPool:
    def __init__(self, nodes: List[str], replica: int):
        self._nodes = nodes
        self._replica = replica
        self._ring_length = len(nodes) * replica
        self._ring = []

        self.make_ring()

    def get_node(self, value: Any) -> 'ConHashNode':
        hash_value = self._hash(value=value)
        node_pos = self.get_node_pos(hash_value=hash_value)

        return self._ring[node_pos]

    def get_node_pos(self, hash_value: int) -> int:
        return self._get_node_pos(hash_value=hash_value, min_pos=0, max_pos=self._ring_length - 1)

    def make_ring(self):
        for index, node in enumerate(self._nodes):
            self._make_ring_for_node(node=node, node_index=index, replica=self._replica)

        self._ring.sort(key=lambda x: x.hash_value)

    def _get_node_pos(self, hash_value: int, min_pos: int, max_pos: int):
        if self._ring[min_pos].hash_value >= hash_value or \
                self._ring[max_pos].hash_value < hash_value or \
                min_pos == max_pos:
            return min_pos

        mid_pos = int((min_pos + max_pos) / 2)
        if self._ring[mid_pos].hash_value < hash_value:
            return self._get_node_pos(hash_value=hash_value, min_pos=mid_pos + 1, max_pos=max_pos)

        return self._get_node_pos(hash_value=hash_value, min_pos=min_pos, max_pos=mid_pos)

    def _make_ring_for_node(self, node: str, node_index: int, replica: int):
        _hash_index = 0
        _start_index = 0

        while _start_index < replica:
            pseudo_name = self._make_pseudo_name(node=node, node_index=node_index, hash_index=_hash_index)
            hash_value = self._hash(pseudo_name)
            _hash_index += 1

            if self._has_hash(hash_value=hash_value):
                continue

            self._ring.append(ConHashNode(hash_value=hash_value, index=node_index))
            _start_index += 1

    def _make_pseudo_name(self, node: str, node_index: int, hash_index: int) -> str:
        _dummy = hash_index * _MAX_32BIT_INTEGER
        return f"{node}:{node_index}:{hash_index}:{_dummy}"

    def _has_hash(self, hash_value: int) -> bool:
        for item in self._ring:
            if item.hash_value == hash_value:
                return True

        return False

    def _hash(self, value: Any) -> int:
        return con_hash.blake2b_hash(value)



from typing import Optional, Type, Tuple, List, Any

from django.conf import settings
from django.db import transaction

from shard.constants import DEFAULT_DATABASE, DATABASE_CONFIG_SHARD_GROUP
from shard.utils.database import get_master_databases_by_shard_group
from shard_static.exceptions import InvalidDatabaseAliasException, NotShardStaticException, NotTransmitException
from shard_static.models import BaseShardStaticModel, BaseStaticTransmitStatus


class Transmitter:
    status_class: BaseStaticTransmitStatus = None

    def __init__(self, shard: str, model_class: Type[BaseShardStaticModel]):
        self.shard = shard
        self.model_class = model_class

        self._validate_model()
        self._validate_database()

    def run(self):
        status = self._get_status()
        last_criterion, items = self.collect(status)
        items = self.pre_transmit(items)

        with transaction.atomic(using=self.shard):
            items = self.transmit(items)
            self.post_transmit(items)
            self._save_status(status, last_criterion)

    def collect(self, status: BaseStaticTransmitStatus) -> Tuple[Any, List[BaseShardStaticModel]]:
        raise NotImplementedError

    def pre_transmit(self, source_items: List[BaseShardStaticModel]) -> List[BaseShardStaticModel]:
        return source_items

    def transmit(self, items: List[BaseShardStaticModel]) -> List[BaseShardStaticModel]:
        raise NotImplementedError

    def post_transmit(self, items: List[BaseShardStaticModel]):
        return items

    def _validate_model(self):
        if not issubclass(self.model_class, BaseShardStaticModel):
            raise NotShardStaticException()

        if not self.model_class.transmit:
            raise NotTransmitException()

    def _validate_database(self):
        if self.shard == DEFAULT_DATABASE:
            raise InvalidDatabaseAliasException()

        if self._get_shard_group_from_database() is None:
            raise InvalidDatabaseAliasException()

        databases = get_master_databases_by_shard_group(shard_group=self.model_class.shard_group)
        if self.shard not in databases:
            raise InvalidDatabaseAliasException()

    def _get_shard_group_from_database(self) -> Optional[str]:
        db_setting = settings.DATABASES.get(self.shard, None)
        if db_setting:
            return db_setting.get(DATABASE_CONFIG_SHARD_GROUP, None)

        return None

    def _get_status(self) -> BaseStaticTransmitStatus:
        status, _ = self.status_class.objects.get_or_create(shard=self.shard, key=self._make_transmit_key())
        return status

    def _save_status(self, status: BaseStaticTransmitStatus, last_criterion) -> BaseStaticTransmitStatus:
        status.criterion = last_criterion
        status.save(using=self.shard)

        return status

    def _make_transmit_key(self) -> str:
        return self.model_class._meta.db_table  # flake8: noqa: W0212 pylint: disable=protected-access



from django.core.management.base import BaseCommand
from pprint import pprint

from shard.utils.shard import get_shard_by_shard_key_and_shard_group


class Command(BaseCommand):
    title = 'Get shard by key and group'

    def add_arguments(self, parser):
        parser.add_argument(
            '--shard_group', '-s',
            type=str,
            dest='shard_group',
            help='Shard group',
            required=True,
        )

        parser.add_argument(
            '--shard_key', '-k',
            type=str,
            dest='shard_key',
            help='Shard Key',
            required=True,
        )

    def handle(self, *args, **options):
        pprint(get_shard_by_shard_key_and_shard_group(options['shard_key'], options['shard_group']))



from shard.mixins import ShardMixin
from shard.providers import pool_provider
from shard.utils.shard_key import get_shard_key_from_instance
from shard.utils.database import get_master_databases_by_shard_group

__all__ = ('get_shard_by_shard_key_and_shard_group', 'get_shard_by_instance', )


def get_shard_by_shard_key_and_shard_group(shard_key: int, shard_group: str) -> str:
    databases = get_master_databases_by_shard_group(shard_group=shard_group)
    shard_index = pool_provider.get_shard_index(shard_group=shard_group, shard_key=shard_key)
    return databases[shard_index]


def get_shard_by_instance(instance: ShardMixin) -> str:
    databases = get_master_databases_by_shard_group(shard_group=instance.shard_group)
    shard_index = pool_provider.get_shard_index(shard_key=get_shard_key_from_instance(instance=instance), shard_group=instance.shard_group)
    return databases[shard_index]



from django.conf import settings

__all__ = ('get_replica_count', )

SHARD_REPLICA_COUNT_SETTING = 'SHARD_REPLICA_COUNT_SETTING'
DEFAULT_REPLICA_COUNT = 512


def get_replica_count(shard_group: str) -> int:
    replica_settings = _get_shard_replica_settings()
    return replica_settings.get(shard_group, DEFAULT_REPLICA_COUNT)


def _get_shard_replica_settings():
    return getattr(settings, SHARD_REPLICA_COUNT_SETTING, {})



# Generated by Django 2.0.2 on 2018-06-07 13:12

from django.db import migrations, models
import shard_static.mixins


class Migration(migrations.Migration):

    dependencies = [
        ('tests', '0002_statictransmitstatus'),
    ]

    operations = [
        migrations.CreateModel(
            name='ShardStaticTransmittableB',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
            ],
            options={
                'abstract': False,
            },
            bases=(shard_static.mixins.ShardStaticMixin, models.Model),
        ),
    ]



from django.db import DatabaseError, connections, transaction
from typing import Dict, List


class ExecuteQueryService:
    @classmethod
    def execute_queries_by_shard(cls, shard: str, queries: List[str]) -> List:
        with connections[shard].cursor() as cursor:
            return cls.execute_queries(cursor, queries)

    @classmethod
    def execute_queries_by_shard_with_transaction(cls, shard: str, queries: List[str]) -> List:
        with transaction.atomic(shard):
            with connections[shard].cursor() as cursor:
                return cls.execute_queries(cursor, queries)

    @classmethod
    def execute_queries(cls, cursor, queries: List[str]) -> List:
        result = []
        for query in queries:
            result.append(cls.execute_query(cursor, query))

        return result

    @classmethod
    def execute_query(cls, cursor, query: str) -> Dict:
        result = {
            'query': query
        }

        try:
            result['result'] = cursor.execute(query)

        except DatabaseError as e:
            result['error'] = str(e)

        if cursor.description:
            result['items'] = cls._parse_result(cursor)

        return result

    @staticmethod
    def _parse_result(cursor):
        columns = [col[0] for col in cursor.description]
        return [
            dict(zip(columns, row))
            for row in cursor.fetchall()
        ]



from shard.mixins import BaseShardMixin


class ShardStaticMixin(BaseShardMixin):
    # If transmit set true, exists source data in default
    # If transmit set false, only exists in shard database
    transmit = True
